from pymbolic.mapper import Collector, IdentityMapper
from pymbolic.primitives import Variable


class IndexPlaceholderBase(object):
    pass


class IndexPlaceholder(IndexPlaceholderBase):
    def __init__(self, expr):
        self.expr = expr

    def __hash__(self):
        return hash(self.expr)

    def __eq__(self, o):
        return (type(o) == IndexPlaceholder) and (self.expr == o.expr)


class LFSIndexPlaceholder(IndexPlaceholderBase):
    def __init__(self, element, restriction, number, context=''):
        self.element = element
        self.restriction = restriction
        self.number = number
        self.context = context

    def __hash__(self):
        return hash((self.element, self.restriction, self.number, self.context))

    def __eq__(self, o):
        return (self.element == o.element) and (self.restriction == o.restriction) and (self.number == o.number) and (self.context == o.context)


class IndexPlaceholderRemoval(IdentityMapper):
    def __init__(self, wrap_in_variable=True, duplicate_inames=False, index_replacement_map={}):
        self.duplicate_inames = duplicate_inames
        self.wrap_in_variable = wrap_in_variable
        self.index_replacement_map = index_replacement_map
        super(IndexPlaceholderRemoval, self).__init__()

    def map_foreign(self, o):
        # How do we map constants here? map_constant was not correct
        if isinstance(o, int) or isinstance(o, float) or isinstance(o, str):
            return o

        # There might be tuples of indices where only one is a placeholder,
        # so we recurse manually into the tuple.
        if isinstance(o, tuple):
            return tuple(self.rec(op) for op in o)

        # We only handle IndexPlaceholder instances from now on
        assert isinstance(o, IndexPlaceholderBase)

        if isinstance(o, LFSIndexPlaceholder):
            context = o.context
            from dune.perftool.loopy.transformer import get_outerloop
            if (self.duplicate_inames) and (o.number != get_outerloop()):
                context = 'dup'

            from dune.perftool.pdelab.basis import lfs_iname
            i = lfs_iname(o.element, o.restriction, count=o.number, context=context)

        if isinstance(o, IndexPlaceholder):
            i = self.index_replacement_map.get(o.expr, o.expr)

        if self.wrap_in_variable and not isinstance(i, Variable):
            return Variable(i)
        else:
            return i

    def map_reduction(self, o):
        from loopy import Reduction
        o.expr = self.rec(o.expr)
        return o


class LFSIndexPlaceholderExtraction(Collector):
    def map_foreign(self, o):
        if isinstance(o, int) or isinstance(o, float):
            return set()
        assert isinstance(o, tuple)
        return set(i for i in o if isinstance(i, LFSIndexPlaceholder))

    def map_reduction(self, o):
        return self.rec(o.expr)
